from django.db import migrations, models
from django.conf import settings

from datetime import datetime

class Migration(migrations.Migration):
    dependencies = [
        ('contribution', '0007_add_missing_fields_not_mssql'),
    ]
    operations = [
            migrations.AlterField(
                model_name='premium',
                name='created_date',
                field=models.DateTimeField(db_column='CreatedDate', default=datetime.now),
            ) 
    ]
        



from contribution.services import premium_updated
from policy.models import Policy
from typing import Optional

import graphene
from contribution.apps import ContributionConfig
from contribution.models import Premium, PremiumMutation
from payer.models import Payer
from policy import models as policy_models
from core.schema import OpenIMISMutation
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import PermissionDenied, ValidationError
from django.utils.translation import gettext as _
from core import datetime
from policy.services import PolicyService
from .services import update_or_create_premium 
import logging

logger = logging.getLogger(__name__)


class PremiumBase:
    """
    This takes most parameters of the Premium with addition of action. This fields allows to force
    """
    id = graphene.Int(required=False, read_only=True)
    uuid = graphene.String(required=False)
    policy_uuid = graphene.String(required=True)
    payer_uuid = graphene.String()
    amount = graphene.Decimal()
    receipt = graphene.String()
    pay_date = graphene.Date()
    pay_type = graphene.String(max_length=1)
    is_offline = graphene.Boolean(required=False)
    is_photo_fee = graphene.Boolean(required=False)
    action = graphene.String(required=False)
    # json_ext = graphene.types.json.JSONString(required=False)


def reset_premium_before_update(premium):
    premium.amount = None
    premium.receipt = None
    premium.policy = None
    premium.payer = None
    premium.pay_date = None
    premium.pay_type = None
    premium.is_photo_fee = None
    premium.is_offline = None
    premium.reporting_id = None


def premium_action(data, user):
    if "client_mutation_id" in data:
        data.pop('client_mutation_id')
    if "client_mutation_label" in data:
        data.pop('client_mutation_label')
    now = datetime.datetime.now()
    data['audit_user_id'] = user.id_for_audit
    data['validity_from'] = now
    policy_uuid = data.pop("policy_uuid") if "policy_uuid" in data else None
    if not policy_uuid:
        raise Exception(_("policy_uuid_required"))
    policy = Policy.filter_queryset(None).filter(uuid=policy_uuid).first()
    if not policy:
        raise Exception(_("policy_uuid_not_found") % (policy_uuid,))
    data["policy"] = policy
    # TODO verify that the user has access to specified payer_id

    # action: enforce, suspend, wait
    action = data.pop("action") if "action" in data else None
    payer_uuid = data.pop("payer_uuid") if "payer_uuid" in data else None
    payer = Payer.filter_queryset().filter(uuid=payer_uuid).first() if payer_uuid else None
    data["payer"] = payer
    premium = Premium(**data)
    # Handle the policy updating
    
    return update_or_create_premium(premium, user, action)


class CreatePremiumMutation(OpenIMISMutation):
    """
    Create a contribution for policy with or without a payer
    """
    _mutation_module = "contribution"
    _mutation_class = "CreatePremiumMutation"

    class Input(PremiumBase, OpenIMISMutation.Input):
        pass

    @classmethod
    def async_mutate(cls, user, **data) -> Optional[str]:
        try:
            if type(user) is AnonymousUser or not user.id:
                raise ValidationError(
                    _("mutation.authentication_required"))
            if not user.has_perms(ContributionConfig.gql_mutation_create_premiums_perms):
                raise PermissionDenied(_("unauthorized"))
            client_mutation_id = data.get("client_mutation_id")
            premium = premium_action(data, user)
            PremiumMutation.object_mutated(user, client_mutation_id=client_mutation_id, premium=premium)
            return None
        except Exception as exc:
            return [{
                'message': _("contribution.mutation.failed_to_create_premium"),
                'detail': str(exc)}
            ]


class UpdatePremiumMutation(OpenIMISMutation):
    """
    Update a contribution for policy with or without a payer
    """
    _mutation_module = "contribution"
    _mutation_class = "UpdatePremiumMutation"

    class Input(PremiumBase, OpenIMISMutation.Input):
        pass

    @classmethod
    def async_mutate(cls, user, **data) -> Optional[str]:
        try:
            if type(user) is AnonymousUser or not user.id:
                raise ValidationError(
                    _("mutation.authentication_required"))
            if not user.has_perms(ContributionConfig.gql_mutation_update_premiums_perms):
                raise PermissionDenied(_("unauthorized"))
            premium_action(data, user)
            return None
        except Exception as exc:
            return [{
                'message': _("contribution.mutation.failed_to_update_premium") %
                {'id': data.get('id') if data else None},
                'detail': str(exc)}
            ]


class DeletePremiumsMutation(OpenIMISMutation):
    """
    Delete one or several Premiums.
    """
    _mutation_module = "contribution"
    _mutation_class = "DeletePremiumsMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(ContributionConfig.gql_mutation_delete_premiums_perms):
            raise PermissionDenied(_("unauthorized"))
        errors = []
        for premium_uuid in data["uuids"]:
            premium = Premium.objects \
                .filter(uuid=premium_uuid) \
                .first()
            if premium is None:
                errors.append({
                    'title': premium_uuid,
                    'list': [{'message': _(
                        "contribution.validation.id_does_not_exist") % {'id': premium_uuid}}]
                })
                continue
            errors += set_premium_deleted(premium)
        if len(errors) == 1:
            errors = errors[0]['list']
        return errors


def set_premium_deleted(premium):
    try:
        premium.delete_history()
        return []
    except Exception as exc:
        logger.debug("Exception when deleting premium %s", premium.uuid, exc_info=exc)
        return {
            'title': premium.uuid,
            'list': [{
                'message': _("contribution.mutation.failed_to_delete_premium") % {'uuid': premium.uuid},
                'detail': premium.uuid}]
        }


def on_policy_mutation(sender, **kwargs):
    errors = []
    if kwargs.get("mutation_class") == 'DeletePoliciesMutation':
        uuids = kwargs['data'].get('uuids', [])
        policies = policy_models.Policy.objects.prefetch_related("premiums").filter(uuid__in=uuids).all()
        for policy in policies:
            for premium in policy.premiums.all():
                errors += set_premium_deleted(premium)
    return errors


def on_premium_mutation(sender, **kwargs):
    uuids = kwargs['data'].get('uuids', [])
    if not uuids:
        uuid = kwargs['data'].get('uuid', None)
        uuids = [uuid] if uuid else []
    if not uuids:
        return []
    impacted_premiums = Premium.objects.filter(uuid__in=uuids).all()
    for premium in impacted_premiums:
        PremiumMutation.objects.update_or_create(premium=premium, mutation_id=kwargs['mutation_log_id'])
    return []



import logging
from enum import Enum
from core.models import filter_validity
from core.datetimes.shared import datetimedelta
from django.db.models import Sum, Q
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
from django.db.transaction import atomic
from insuree.models import Insuree, Family, InsureePolicy
from location.apps import LocationConfig
from location.models import Location
from policy.models import Policy
from policy.services import policy_status_premium_paid

from .models import Premium, PayTypeChoices

logger = logging.getLogger(__name__)

# A fake family is used for funding
FUNDING_CHF_ID = "999999999"


class ByPolicyPremiumsAmountService(object):
    def __init__(self, user):
        self.user = user

    def request(self, policy_id):
        return (
            Premium.objects.filter(policy_id=policy_id)
            .exclude(is_photo_fee=True)
            .aggregate(Sum("amount"))["amount__sum"]
        )


def last_date_for_payment(policy_id):
    policy = Policy.objects.get(id=policy_id)
    has_cycle = policy.product.has_cycle()

    if policy.stage == "N":
        grace_period = policy.product.grace_period_enrolment
    elif policy.stage == "R":
        grace_period = policy.product.grace_period_renewal
    else:
        logger.error(
            "policy stage should be either N or R, policy %s has %s",
            policy_id,
            policy.stage,
        )
        raise Exception("policy stage should be either N or R")

    waiting_period = policy.product.grace_period_payment

    if has_cycle:
        # Calculate on fixed cycle
        start_date = policy.start_date

        last_date = start_date + datetimedelta(months=grace_period)
    else:
        # Calculate on Free Cycle
        if policy.stage == "N":
            last_date = policy.enroll_date + datetimedelta(months=waiting_period)
        else:
            last_date = (
                policy.expiry_date
                + datetimedelta(days=1)
                + datetimedelta(months=waiting_period)
            )

    return last_date - datetimedelta(days=1)


def can_payer_fund_product(payer, product):
    if not product.location:
        return True
    elif product.location == payer.location:
        return True
    elif product.location.type == "R" and payer.location.parent == product.location:
        return True
    elif product.location.type == "D" and payer.location == product.location.parent:
        return True

    return False


@atomic
def add_fund(payer, product, pay_date, amount, receipt, audit_user_id, is_offline):
    # We create fake locations for fundings
    fundings = []
    funding_parent = None  # Top Funding has no parent, then the loop will chain them
    for level in LocationConfig.location_types:
        level_funding, funding_created = Location.objects.get_or_create(
            code=f"F{level}",
            name="Funding",
            parent=funding_parent,
            type=level,
            defaults=dict(audit_user_id=audit_user_id),
        )
        funding_parent = level_funding
        fundings.append(level_funding)
        if funding_created:
            logger.warning("Created funding at level %s", level)

    if product.validity_to is not None:
        raise ValueError("Product has to be valid")

    if not can_payer_fund_product(payer, product):
        raise ValueError("Payer and product locations are incompatible")

    # TODO check and/or document premium_adult
    product_value = product.lump_sum or product.premium_adult

    # Check if the family with CHFID exists
    # Original procedure here has a strange and useless join on isnull(,0), ignoring
    family = (
        Family.objects.filter(validity_to__isnull=True)
        .filter(head_insuree__chf_id=FUNDING_CHF_ID)
        .filter(location_id=product.location_id)
        .first()
    )

    if not family:
        insuree = Insuree.objects.create(
            family=family,
            chf_id=FUNDING_CHF_ID,
            last_name="Funding",
            other_names="Funding",
            gender=None,
            marital=None,
            head=True,
            card_issued=False,
            dob=pay_date,
            audit_user_id=audit_user_id,
            offline=is_offline,
        )
        family = Family.objects.create(
            head_insuree=insuree,
            location_id=product.location_id,
            poverty=False,
            is_offline=is_offline,
            audit_user_id=audit_user_id,
        )

    from core import datetimedelta

    policy = Policy.objects.create(
        family=family,
        enroll_date=pay_date,
        start_date=pay_date,
        effective_date=pay_date,
        expiry_date=datetimedelta(months=product.insurance_period).add_to_date(
            pay_date
        ),
        status=Policy.STATUS_ACTIVE,
        value=product_value,
        product=product,
        officer_id=None,
        audit_user_id=audit_user_id,
        offline=is_offline,
    )

    InsureePolicy.objects.create(
        insuree=family.head_insuree,
        policy=policy,
        enrollment_date=policy.enroll_date,
        start_date=policy.start_date,
        effective_date=policy.effective_date,
        expiry_date=policy.expiry_date,
        audit_user_id=audit_user_id,
        offline=is_offline,
    )

    return Premium.objects.create(
        policy=policy,
        payer_id=payer.id,
        amount=amount,
        receipt=receipt,
        pay_date=pay_date,
        pay_type=PayTypeChoices.FUNDING,
        is_offline=is_offline,
        audit_user_id=audit_user_id,
    )
from django.db.models import Q


class PremiumUpdateActionEnum(Enum):
    SUSPEND = "SUSPEND"
    ENFORCE = "ENFORCE"
    WAIT = "WAIT"


def premium_updated(premium, action=None):
    """
    if the contribution is lower than the policy value, action can override it or suspend the policy
    if it is right or too much, just activate it (enforce is still expected but just a warning)
    """
    policy = premium.policy
    policy.save_history()

    if action == PremiumUpdateActionEnum.SUSPEND.value:
        policy.status = Policy.STATUS_SUSPENDED
        policy.save()
        return

    policy_balance = policy.value - premium.other_premiums()
    
    if premium.amount  == policy_balance:
        policy_status_premium_paid(
            policy,
            premium.pay_date
            if premium.pay_date > policy.start_date
            else policy.start_date,
        )
    elif premium.amount < policy_balance:
        # suspend already handledpremium
        if action == PremiumUpdateActionEnum.ENFORCE.value:
            policy_status_premium_paid(policy, premium.pay_date)
        # otherwise, just leave the policy unchanged
    elif premium.amount > policy_balance:
        if action != PremiumUpdateActionEnum.ENFORCE.value:
            logger.warning("action on premiums larger than the policy value")
        policy_status_premium_paid(policy, premium.pay_date)
    else:
        logger.warning(
            "The comparison between premium amount %s and policy value %s failed",
            premium.amount,
            policy.value,
        )
        raise Exception("Invalid combination or premium and policy amounts")

    if policy.status is not None and (
        policy.effective_date == premium.pay_date
        or policy.effective_date == policy.start_date
    ):
        # Enforcing policy
        if policy.offline or not premium.is_offline:
            policy.save()
        if policy.status == Policy.STATUS_ACTIVE:
            _update_policy_insurees(policy)
    elif policy.effective_date:
        _activate_insurees(policy, premium.pay_date)


def _update_policy_insurees(policy):
    policy.insuree_policies.filter(validity_to__isnull=True).update(
        effective_date=policy.effective_date,
        start_date=policy.start_date,
        expiry_date=policy.expiry_date,
    )


def _activate_insurees(policy, pay_date):
    policy.insuree_policies.filter(validity_to__isnull=True).update(
        effective_date=pay_date,
    )


def check_unique_premium_receipt_code_within_product(code, policy_uuid = None, policy = None):
    from .models import Premium

    if not policy:
        if not policy_uuid:
            return [{"message": "missing Policy"}]
        policy = Policy.objects.select_related('product').filter(uuid=policy_uuid, validity_to__isnull=True).first()
    exists = Premium.objects.filter(policy__product=policy.product, receipt=code, validity_to__isnull=True).exists()
    if exists:
        return [{"message": "Premium code %s already exists" % code}]
    return []


def update_or_create_premium(premium, user, action=None):
    existing_premium = Premium.objects.filter(*filter_validity(), Q(Q(uuid=premium.uuid) | Q(id=premium.id))).first()
    if existing_premium:
        return update_premium(existing_premium, premium, user, action)
    else:  
        return create_premium(premium, user, action)


def update_premium(existing_premium, premium, user, action = None):
    if existing_premium.receipt != premium.receipt:
        if check_unique_premium_receipt_code_within_product(code=premium.receipt, policy=premium.policy):
            raise ValidationError(_("mutation.code_already_taken"))
    existing_premium.save_history()
    premium.id = existing_premium.id
    premium.save()
    # Handle the policy updating
    premium_updated(premium, action)
    return premium


def create_premium(premium, user, action = None):
    if check_unique_premium_receipt_code_within_product(code=premium.receipt, policy=premium.policy):
        raise ValidationError(_("mutation.code_already_taken"))
    premium.save()
    # Handle the policy updating
    premium_updated(premium, action)
    return premium


# Generated by Django 3.2.18 on 2023-05-12 12:15
import core.datetimes.ad_datetime
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contribution', '0004_auto_20230718_1155'),
    ]

    operations = [
            migrations.RemoveField(
            model_name='premium',
            name='row_id',
        ),
    ]
    if settings.MSSQL:
        operations.append(
            migrations.AlterField(
                model_name='premium',
                name='created_date',
                field=models.DateTimeField(db_column='CreatedDate', default=core.datetimes.ad_datetime.AdDatetime.now),
            ))
        




# Generated by Django 3.2.19 on 2023-07-18 11:55
from django.conf import settings
from django.db import migrations, models

from contribution.utils import AddFieldPostgres


class Migration(migrations.Migration):
    dependencies = [
        ('contribution', '0003_alter_premium_options'),
    ]

    # For MSSQL These changes were added through raw sql, to keep consistency with existing databases it couldn't be
    # changed in postgres sql script.
    operations = []
    if not settings.MSSQL:
        operations = [
            AddFieldPostgres(
                model_name='premium',
                name='all_details_commission_report',
                field=models.DateTimeField(blank=True, db_column='AllDetailsCommissionReport', null=True),
            ),
            AddFieldPostgres(
                model_name='premium',
                name='overview_commission_report',
                field=models.DateTimeField(blank=True, db_column='OverviewCommissionReport', null=True),
            ),
            AddFieldPostgres(
                model_name='premium',
                name='reporting_commission_id',
                field=models.IntegerField(blank=True, db_column='ReportingCommissionID', null=True),
            )
        ]



# Generated by Django 3.0.3 on 2021-01-19 10:08

import core.fields
import datetime
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0009_mutationlog_client_mutation_details'),
    ]

    operations = [
        migrations.CreateModel(
            name='Premium',
            fields=[
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now)),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('id', models.AutoField(db_column='PremiumId', primary_key=True, serialize=False)),
                ('uuid', models.CharField(db_column='PremiumUUID', default=uuid.uuid4, max_length=36, unique=True)),
                ('amount', models.DecimalField(db_column='Amount', decimal_places=2, max_digits=18)),
                ('receipt', models.CharField(db_column='Receipt', max_length=50)),
                ('pay_date', core.fields.DateField(db_column='PayDate')),
                ('pay_type', models.CharField(choices=[('B', 'Bank transfer'), ('C', 'Cash'), ('M', 'Mobile phone'), ('F', 'Funding')], db_column='PayType', max_length=1)),
                ('is_photo_fee', models.BooleanField(blank=True, db_column='isPhotoFee', null=True)),
                ('is_offline', models.BooleanField(blank=True, db_column='isOffline', null=True)),
                ('reporting_id', models.IntegerField(blank=True, db_column='ReportingId', null=True)),
                ('audit_user_id', models.IntegerField(db_column='AuditUserID')),
                ('row_id', models.BinaryField(blank=True, db_column='RowID', null=True)),
                ('source', models.CharField(db_column="Source", max_length=50, blank=True, null=True)),
                ('source_version', models.CharField(db_column="SourceVersion", max_length=15, blank=True, null=True)),
            ],
            options={
                'db_table': 'tblPremium',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='PremiumMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='premiums', to='core.MutationLog')),
                ('premium', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='contribution.Premium')),
            ],
            options={
                'db_table': 'contribution_PremiumMutation',
                'managed': True,
            },
        ),
    ]



from django.db import migrations, models

from core.datetimes.ad_datetime import datetime


class Migration(migrations.Migration):
    dependencies = [
        ('contribution', '0006_add_source_field_to_contribution'),
    ]
    #this migration was useless, better remove it and correct in next migration
    operations = [
    ]



from django.db import migrations, models


class AddFieldPostgres(migrations.AddField):
    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        if schema_editor.connection.vendor == 'postgresql':
            super().database_forwards(app_label, schema_editor, from_state, to_state)

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        if schema_editor.connection.vendor == 'postgresql':
            super().database_backwards(app_label, schema_editor, from_state, to_state)

    def describe(self):
        # This is used to describe what the operation does in console output.
        return "Wrapper for AddField that works only for postgres database engine."



from django.db import migrations, models
from django.db.utils import ProgrammingError, OperationalError

def check_field_exists(apps, schema_editor, field_name):
    Premium = apps.get_model('contribution', 'Premium')
    db_alias = schema_editor.connection.alias
    
    try:
        Premium.objects.using(db_alias).values(field_name).first()
        return True
    except (ProgrammingError, OperationalError):
        return False

def add_field_if_not_exists(apps, schema_editor, field_name, field):
    if not check_field_exists(apps, schema_editor, field_name):
        schema_editor.add_field(
            model=apps.get_model('contribution', 'Premium'),
            field=field
        )

class Migration(migrations.Migration):

    dependencies = [
        ('contribution', '0005_alter_created_date'),
    ]

    operations = [
        migrations.RunPython(
            lambda apps, schema_editor: add_field_if_not_exists(
                apps, 
                schema_editor, 
                'source',
                models.CharField(db_column="Source", max_length=50, blank=True, null=True)
            ),
            reverse_code=migrations.RunPython.noop
        ),
        migrations.RunPython(
            lambda apps, schema_editor: add_field_if_not_exists(
                apps, 
                schema_editor, 
                'source_version',
                models.CharField(db_column="SourceVersion", max_length=15, blank=True, null=True)
            ),
            reverse_code=migrations.RunPython.noop
        ),
    ]


import graphene
from django.db.models import Q, Sum, F
import graphene_django_optimizer as gql_optimizer

from .apps import ContributionConfig
from location.apps import LocationConfig
from django.utils.translation import gettext as _
from core.schema import signal_mutation_module_before_mutating, OrderedDjangoFilterConnectionField, filter_validity
from core.services import wait_for_mutation
# We do need all queries and mutations in the namespace here.
from .gql_queries import *  # lgtm [py/polluting-import]
from .gql_mutations import *  # lgtm [py/polluting-import]
from .services import check_unique_premium_receipt_code_within_product
from graphql.language.ast import Field as ASTField
def ast_to_dict(node, with_location=False):
    if isinstance(node, list):
        return [ast_to_dict(item, with_location) for item in node]
    if node  and hasattr(node, '_fields'):
        res = {key: ast_to_dict(getattr(node, key), with_location)
               for key in node._fields if key != 'loc' }
        if with_location:
            loc = node.loc
            if loc:
                res['loc'] = dict(start=loc.start, end=loc.end)
        
        res['kind'] = node.kind if hasattr(node, 'kind') else 'Field'
        return res
    return node

def collect_fields(node, fragments):
    """Recursively collects fields from the AST
    Args:
        node (dict): A node in the AST
        fragments (dict): Fragment definitions
    Returns:
        A dict mapping each field found, along with their sub fields.
        {'name': {},
         'sentimentsPerLanguage': {'id': {},
                                   'name': {},
                                   'totalSentiments': {}},
         'slug': {}}
    """

    field = {}

    if node.get('selection_set'):
        for leaf in node['selection_set']['selections']:
            if leaf['kind'] == 'Field':
                field.update({
                    leaf['name']['value']: collect_fields(leaf, fragments)
                })
            elif leaf['kind'] == 'FragmentSpread':
                field.update(collect_fields(fragments[leaf['name']['value']],
                                            fragments))
    return field


def get_fields(info):
    """A convenience function to call collect_fields with info
    Args:
        info (ResolveInfo)
    Returns:
        dict: Returned from collect_fields
    """

    fragments = {}
    node = ast_to_dict(info.field_asts[0])

    for name, value in info.fragments.items():
        fragments[name] = ast_to_dict(value)

    return collect_fields(node, fragments)

class Query(graphene.ObjectType):
    premiums = OrderedDjangoFilterConnectionField(
        PremiumGQLType,
        payer_id=graphene.ID(),
        client_mutation_id=graphene.String(),
        show_history=graphene.Boolean(),
        parent_location=graphene.String(),
        parent_location_level=graphene.Int(),
        orderBy=graphene.List(of_type=graphene.String),
    )
    premiums_by_policies = OrderedDjangoFilterConnectionField(
        PremiumGQLType,
        policy_uuids=graphene.List(graphene.String, required=True),
        orderBy=graphene.List(of_type=graphene.String),
    )
    validate_premium_code = graphene.Field(
        graphene.Boolean,
        code=graphene.String(required=True),
        policy_uuid=graphene.String(required=True),
        description="Checks that the specified premium code is unique for a given policy."
    )

    def resolve_premiums(self, info, **kwargs):
        fields = get_fields(info)
        queryset = Premium.objects
        if not info.context.user.has_perms(ContributionConfig.gql_query_premiums_perms):
            raise PermissionDenied(_("unauthorized"))
        filters = []
        client_mutation_id = kwargs.get("client_mutation_id", None)
        payer_id = kwargs.get("payer_id", None)
        if payer_id:
            filters.append(Q(payer__id=payer_id))
        if client_mutation_id:
            wait_for_mutation(client_mutation_id)
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))
        show_history = kwargs.get('show_history', False)
        if not show_history and not kwargs.get('uuid', None):
            filters += filter_validity(**kwargs)
        parent_location = kwargs.get('parent_location')
        if parent_location is not None:
            parent_location_level = kwargs.get('parent_location_level')
            if parent_location_level is None:
                raise NotImplementedError("Missing parentLocationLevel argument when filtering on parentLocation")
            f = "uuid"
            for i in range(len(LocationConfig.location_types) - parent_location_level - 1):
                f = "parent__" + f
            family_location = "policy__family__location__" + f
            filters.append(Q(**{family_location: parent_location}))
        if 'otherPremiums' in fields['edges']['node']:
            queryset = queryset.annotate(other_premiums = Sum('policy__premiums__amount', filter = Q(~Q(policy__premiums__id=F('id')) & Q(*filter_validity(prefix='policy__premiums__'),policy__premiums__is_photo_fee=False))))
        return gql_optimizer.query(queryset.filter(*filters).all(), info)

    def resolve_premiums_by_policies(self, info, **kwargs):
        if not info.context.user.has_perms(ContributionConfig.gql_query_premiums_perms):
            raise PermissionDenied(_("unauthorized"))
        policies = policy_models.Policy.objects.values_list('id').filter(Q(uuid__in=kwargs.get('policy_uuids')))
        return Premium.objects.filter(Q(policy_id__in=policies), *filter_validity(**kwargs))

    def resolve_validate_premium_code(self, info, **kwargs):
        if not info.context.user.has_perms(ContributionConfig.gql_query_premiums_perms):
            raise PermissionDenied(_("unauthorized"))
        errors = check_unique_premium_receipt_code_within_product(code=kwargs['code'],
                                                                  policy_uuid=kwargs['policy_uuid'])
        return False if errors else True


def set_premium_deleted(premium):
    try:
        premium.delete_history()
        return []
    except Exception as exc:
        return {
            'title': premium.uuid,
            'list': [{
                'message': _("premium.mutation.failed_to_delete_premium") % {'premium': str(premium)},
                'detail': premium.uuid
            }]
        }


class Mutation(graphene.ObjectType):
    delete_premium = DeletePremiumsMutation.Field()
    create_premium = CreatePremiumMutation.Field()
    update_premium = UpdatePremiumMutation.Field()


def bind_signals():
    signal_mutation_module_before_mutating["policy"].connect(on_policy_mutation)
    signal_mutation_module_before_mutating["contribution"].connect(on_premium_mutation)



import uuid

from contribution.models import Payer, Premium
from payer.test_helpers import create_test_payer


def create_test_premium(policy_id, with_payer=True, custom_props=None):
    payer = create_test_payer() if with_payer else None
    if custom_props is None:
        custom_props = {}
    else:
        custom_props = {k: v for k, v in custom_props.items() if hasattr(Premium, k)}

    premium = Premium.objects.create(
        **{
            "policy_id": policy_id,
            "payer_id": payer.id if payer else None,
            "amount": 1000,
            "receipt": "Test receipt",
            "pay_date": "2019-01-01",
            "pay_type": Payer.PAYER_TYPE_OTHER,
            "validity_from": "2019-01-01",
            "audit_user_id": -1,
            **custom_props
        }
    )


    return premium



from contribution.reports import premium_collection, payment_category_overview, contributions_distribution
from contribution.reports.contributions_distribution import contributions_distribution_query
from contribution.reports.payment_category_overview import payment_category_overview_query
from contribution.reports.premium_collection import premium_collection_query

report_definitions = [
    {
        "name": "premium_collection",
        "engine": 0,
        "default_report": premium_collection.template,
        "description": "Premium collection",
        "module": "contribution",
        "python_query": premium_collection_query,
        "permission": ["131204"],
    },
    {
        "name": "payment_category_overview",
        "engine": 0,
        "default_report": payment_category_overview.template,
        "description": "Payment category overview",
        "module": "contribution",
        "python_query": payment_category_overview_query,
        "permission": ["131211"],
    },
    {
        "name": "contributions_distribution",
        "engine": 0,
        "default_report": contributions_distribution.template,
        "description": "Contributions distribution",
        "module": "contribution",
        "python_query": contributions_distribution_query,
        "permission": ["131206"],
    },
]



import uuid

from core import fields
from core import models as core_models
from django.db import models
from django.utils.translation import gettext_lazy

from core.datetimes.ad_datetime import datetime
from policy.models import Policy
from payer.models import Payer


class PayTypeChoices(models.TextChoices):
    BANK_TRANSFER = "B", gettext_lazy("Bank transfer")
    CASH = "C", gettext_lazy("Cash")
    MOBILE = "M", gettext_lazy("Mobile phone")
    FUNDING = "F", gettext_lazy("Funding")


class Premium(core_models.VersionedModel):
    id = models.AutoField(db_column="PremiumId", primary_key=True)
    uuid = models.CharField(
        db_column="PremiumUUID", max_length=36, default=uuid.uuid4, unique=True
    )
    policy = models.ForeignKey(
        Policy, models.DO_NOTHING, db_column="PolicyID", related_name="premiums"
    )
    payer = models.ForeignKey(
        Payer,
        models.DO_NOTHING,
        db_column="PayerID",
        blank=True,
        null=True,
        related_name="premiums",
    )
    amount = models.DecimalField(
        db_column="Amount", max_digits=18, decimal_places=2)
    receipt = models.CharField(db_column="Receipt", max_length=50)
    pay_date = fields.DateField(db_column="PayDate")
    pay_type = models.CharField(
        db_column="PayType", max_length=1
    )  # , choices=PayTypeChoices.choices
    is_photo_fee = models.BooleanField(
        db_column="isPhotoFee", blank=True, null=True, default=False
    )
    is_offline = models.BooleanField(
        db_column="isOffline", blank=True, null=True, default=False
    )
    reporting_id = models.IntegerField(
        db_column="ReportingId", blank=True, null=True)
    reporting_commission_id = models.IntegerField(
        db_column="ReportingCommissionID", blank=True, null=True)
    overview_commission_report = models.DateTimeField(
        db_column="OverviewCommissionReport", blank=True, null=True)
    all_details_commission_report = models.DateTimeField(
        db_column="AllDetailsCommissionReport", blank=True, null=True)
    source = models.CharField(
        db_column="Source", max_length=50, blank=True, null=True)
    source_version = models.CharField(
        db_column="SourceVersion", max_length=15, blank=True, null=True)

    audit_user_id = models.IntegerField(db_column="AuditUserID")
    created_date = models.DateTimeField(
        db_column="CreatedDate", default=datetime.now)
    #rowid = models.TextField(db_column='RowID', blank=True, null=True)
    
    def other_premiums(self):
        return self.policy.premiums.aggregate(
            other_premiums = models.Sum(
                'amount',
                filter = models.Q(~models.Q(id=self.id) & models.Q(*core_models.filter_validity(prefix=''),is_photo_fee=False))
            )
        )['other_premiums'] or 0
    class Meta:
        managed = True
        db_table = 'tblPremium'


class PremiumMutation(core_models.UUIDModel, core_models.ObjectMutation):
    premium = models.ForeignKey(
        Premium, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(
        core_models.MutationLog, models.DO_NOTHING, related_name='premiums')

    class Meta:
        managed = True
        db_table = "contribution_PremiumMutation"



import graphene
from graphene_django import DjangoObjectType

from .apps import ContributionConfig
from .models import Premium, PremiumMutation
from core import prefix_filterset, ExtendedConnection
from policy.schema import PolicyGQLType
from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext as _


class PremiumGQLType(DjangoObjectType):
    client_mutation_id = graphene.String()
    other_premiums = graphene.Float(source="other_premiums")

    class Meta:
        model = Premium
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "uuid": ["exact"],
            "amount": ["exact", "lt", "lte", "gt", "gte"],
            "pay_date": ["exact", "lt", "lte", "gt", "gte"],
            "pay_type": ["exact"],
            "is_photo_fee": ["exact"],
            "receipt": ["exact", "icontains"],
            **prefix_filterset("policy__", PolicyGQLType._meta.filter_fields)
        }
        connection_class = ExtendedConnection

    def resolve_client_mutation_id(self, info):
        if not info.context.user.has_perms(ContributionConfig.gql_query_premiums_perms):
            raise PermissionDenied(_("unauthorized"))
        premium_mutation = self.mutations.select_related(
            'mutation').filter(mutation__status=0).first()
        return premium_mutation.mutation.client_mutation_id if premium_mutation else None


class PremiumMutationGQLType(DjangoObjectType):
    class Meta:
        model = PremiumMutation



import base64
import json
from dataclasses import dataclass

from core.models import User, filter_validity
from core.test_helpers import create_test_interactive_user
from django.conf import settings
from graphene_django.utils.testing import GraphQLTestCase
from graphql_jwt.shortcuts import get_token
from location.models import Location
from location.test_helpers import create_test_location, assign_user_districts
from rest_framework import status
from insuree.test_helpers import create_test_insuree
from policy.test_helpers import create_test_policy
from contribution.models import Premium, PayTypeChoices

from location.test_helpers import create_test_location, create_test_health_facility, create_test_village
from payer.models import Payer
from product.models import Product
import datetime

# from openIMIS import schema


@dataclass
class DummyContext:
    """ Just because we need a context to generate. """
    user: User


class ContributionGQLTestCase(GraphQLTestCase):
    GRAPHQL_URL = f'/{settings.SITE_ROOT()}graphql'
    # This is required by some version of graphene but is never used. It should be set to the schema but the import
    # is shown as an error in the IDE, so leaving it as True.
    GRAPHQL_SCHEMA = True
    admin_user = None
    ca_user = None
    ca_token = None
    test_village = None
    test_insuree = None
    policy = None
    product = None
    payer = None
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.test_village = create_test_village()
        cls.test_insuree = create_test_insuree(with_family=True, is_head=True, custom_props={'current_village':cls.test_village}, family_custom_props={'location':cls.test_village})
        cls.admin_user = create_test_interactive_user(username="testLocationAdmin")
        cls.admin_token = get_token(cls.admin_user, DummyContext(user=cls.admin_user))
        cls.ca_user = create_test_interactive_user(username="testLocationNoRight", roles=[9])
        cls.ca_token = get_token(cls.ca_user, DummyContext(user=cls.ca_user))
        cls.admin_dist_user = create_test_interactive_user(username="testLocationDist")
        assign_user_districts(cls.admin_dist_user, ["R1D1", "R2D1", "R2D2", "R2D1", cls.test_village.parent.parent.code])
        cls.admin_dist_token = get_token(cls.admin_dist_user, DummyContext(user=cls.admin_dist_user))
        cls.payer = Payer.objects.filter(*filter_validity()).first()
        cls.product = Product.objects.filter(*filter_validity()).first()
        cls.policy = create_test_policy(cls.product, cls.test_insuree, custom_props={'value':1000}, link=True, valid=True)


        
    def test_add_funding(self):
      
        response = self.query(
      f'''
    mutation {{
      createPremium(
        input: {{
          uuid: "94a07513-87b9-469e-bb73-58eb717fee05"
          clientMutationId: "94a07513-87b9-469e-bb73-58eb717fee05"
          clientMutationLabel: "Create contribution"
          receipt: "ghfjgfhj"
          payDate: "2023-12-13"
          payType: "C"
          isPhotoFee: false
          amount: "4200"
          policyUuid: "{self.policy.uuid}"
              }}
      ) {{
        clientMutationId
        internalId
      }}
    }}

      ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
        )

        self.assertEquals(response.status_code, status.HTTP_200_OK)
        content = json.loads(response.content)

        self.assertResponseNoErrors(response)
        #wait 
        
        response = self.query('''
        
        {
        mutationLogs(clientMutationId: "94a07513-87b9-469e-bb73-58eb717fee05")
        {
            
        pageInfo { hasNextPage, hasPreviousPage, startCursor, endCursor}
        edges
        {
        node
        {
            id,status,error,clientMutationId,clientMutationLabel,clientMutationDetails,requestDateTime,jsonExt
        }
        }
        }
        }
        
        ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"})
        
        self.assertEquals(response.status_code, status.HTTP_200_OK)
        content = json.loads(response.content)
        self.assertResponseNoErrors(response)
        premium = Premium.objects.filter(uuid = "94a07513-87b9-469e-bb73-58eb717fee05",*filter_validity()).first()
        self.assertIsNotNone(premium)
        self.assertEquals(premium.amount, 4200)
        #modify premium
        
        response = self.query(
      f'''
    mutation {{
      updatePremium(
        input: {{
          uuid: "94a07513-87b9-469e-bb73-58eb717fee05"
          clientMutationId: "94a07513-87b9-469e-bb73-58eb717fee32"
          clientMutationLabel: "Create contribution"
          receipt: "ghfjgfhj"
          payDate: "2023-12-13"
          payType: "C"
          isPhotoFee: false
          amount: "4400"
          policyUuid: "{self.policy.uuid}"
              }}
      ) {{
        clientMutationId
        internalId
      }}
    }}

      ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
        )
        self.assertEquals(response.status_code, status.HTTP_200_OK)
        content = json.loads(response.content)
        self.assertResponseNoErrors(response)
        premium = Premium.objects.filter(uuid = "94a07513-87b9-469e-bb73-58eb717fee05",*filter_validity()).first()
        self.assertIsNotNone(premium)
        self.assertEquals(premium.amount, 4400)

        
    def test_query_premium(self):
        premium = Premium.objects.create(**{
              'payer':self.payer,
              'pay_date':datetime.datetime.now(),
              'amount':"5000",
              'receipt':"test premium 1",
              'policy':self.policy,
              'pay_type':PayTypeChoices.BANK_TRANSFER,
              'audit_user_id':self.admin_user.id_for_audit
          }
          )
      
        premium = Premium.objects.create(**{
                'payer':self.payer,
                'pay_date':datetime.datetime.now(),
                'amount':"10000",
                'receipt':"test premium 2",
                'policy':self.policy,
                'pay_type':PayTypeChoices.BANK_TRANSFER,
                'audit_user_id':self.admin_user.id_for_audit
            }
            )
        response = self.query(
          f'''
           {{ 
            premiums(uuid: "{premium.uuid}")
            {{
              
          pageInfo {{ hasNextPage, hasPreviousPage, startCursor, endCursor}}
          edges
          {{
            node
            {{
              id,uuid,payDate,amount,payType,receipt,isPhotoFee,validityTo,clientMutationId,otherPremiums,payer{{id, uuid, name}},policy{{id, uuid, startDate, product{{name, code}}, expiryDate, value, sumPremiums, family{{id, uuid, headInsuree{{chfId, lastName, otherNames, dob}}}}
            }}
          }}
            }}
          }}
          }}
          '''
          ,headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"})
        self.assertEquals(response.status_code, status.HTTP_200_OK)
        content = json.loads(response.content)

        self.assertResponseNoErrors(response)

        
    def test_query_premiums(self):
        response = self.query(
        '''
        {
          premiums(first: 10,orderBy: ["-payDate"])
          {
            totalCount
            
        pageInfo { hasNextPage, hasPreviousPage, startCursor, endCursor}
        edges
        {
          node
          {
            id,uuid,payDate,amount,payType,receipt,isPhotoFee,clientMutationId,validityTo,payer{id, uuid, name}
          }
        }
          }
        }
        ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
            variables={ 'first':10, 'payerId':self.payer.uuid},
        )

        self.assertEquals(response.status_code, status.HTTP_200_OK)
        content = json.loads(response.content)

        self.assertResponseNoErrors(response)

        







# Generated by Django 3.2.18 on 2023-05-12 12:15

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contribution', '0002_add_premium_fields'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='premium',
            options={'managed': True},
        ),
    ]



# Generated by Django 3.2.18 on 2023-05-12 12:15

import core.datetimes.ad_datetime
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contribution', '0001_initial'),
        ('payer', '0001_initial'),
        ('policy', '0004_add_medical_oficer_reading_rights'),
    ]

    operations = [
        migrations.AddField(
            model_name='premium',
            name='created_date',
            field=models.DateTimeField(db_column='CreatedDate', default=core.datetimes.ad_datetime.AdDatetime.now),
        ),
        migrations.AddField(
            model_name='premium',
            name='payer',
            field=models.ForeignKey(blank=True, db_column='PayerID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='premiums', to='payer.payer'),
        ),
        migrations.AddField(
            model_name='premium',
            name='policy',
            field=models.ForeignKey(db_column='PolicyID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='premiums', to='policy.policy'),
        ),
        migrations.AlterField(
            model_name='premium',
            name='is_offline',
            field=models.BooleanField(blank=True, db_column='isOffline', default=False, null=True),
        ),
        migrations.AlterField(
            model_name='premium',
            name='is_photo_fee',
            field=models.BooleanField(blank=True, db_column='isPhotoFee', default=False, null=True),
        ),
        migrations.AlterField(
            model_name='premium',
            name='pay_type',
            field=models.CharField(db_column='PayType', max_length=1),
        ),
    ]
